// Dummy file for creating a first mrc-mini application

#include "mrsrc.h"

const mrsrc::rsrc_imp gResourceIndex[0] = {};
const char gResourceData[] = "\0\0\0\0";
const char gResourceName[] = "\0\0\0\0";


